/*
 This file is part of the digitalSTROM dsm-api library.

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM dsm-api library is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 digitalSTROM dsm-api library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with digitalSTROM dsm-api library. If not, see <http://www.gnu.org/licenses/>.
 */

#include "dsm-api.h"
#include "uthash.h"
#include <digitalSTROM/ds485-client-interface.h>
#include "dsm-macros.h"
#include <stdlib.h>
#include <errno.h>
#include <pthread.h>

typedef struct {
    ds485ClientHandle_t client;
    struct callback_item *callback_list;
    BusState_callback_t bus_state_callback;
    void *bus_state_arg;
    BusChange_callback_t bus_change_callback;
    void *bus_change_arg;
    pthread_mutex_t mutex_callback;
    pthread_mutex_t mutex;
    pthread_cond_t condition;
    void *sync_data;
} dsm_api_data_t;

static int dsm_api_ds485_message_callback(void *handle, const ds485_container_t *container);
static int dsm_api_ds485_bus_state_callback(void *handle, const bus_state_t state);
static int dsm_api_ds485_bus_change_callback(void *handle, dsid_t *id, int flags);

typedef void (*callback_function_t)();

struct callback_item {
    uint32_t id;
    callback_function_t callback;
    void *arg;
    UT_hash_handle hh;
};

static struct callback_item *callback_list = NULL;

int DsmApiLibraryVersion(void)
{
    return API_VERSION;
}

DsmApiHandle_t DsmApiInitialize(void)
{
    dsm_api_data_t *data = (dsm_api_data_t*)calloc(sizeof(dsm_api_data_t), sizeof(uint8_t));

    if (data == NULL)
    {
        return NULL;
    }

    data->callback_list = NULL;
    data->bus_change_callback = NULL;
    data->bus_state_callback = NULL;
    data->sync_data = NULL;

    pthread_mutex_init(&data->mutex_callback, NULL);
    pthread_mutex_init(&data->mutex, NULL);
    pthread_cond_init(&data->condition, NULL);

    return data;
}

int DsmApiOpen(DsmApiHandle_t handle, const char *connection, int flags)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    data->client = ds485_client_open(connection, flags);
    if (data->client == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    /* register message callback with lower layer */
    ds485_client_set_receive_callback(data->client, &dsm_api_ds485_message_callback, data);
    ds485_client_set_state_callback(data->client, &dsm_api_ds485_bus_state_callback, data);
    ds485_client_set_bus_change_callback(data->client, &dsm_api_ds485_bus_change_callback, data);

    return 0;
}

int DsmApiClose(DsmApiHandle_t handle)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    struct callback_item *item;

    /* clear message callback with lower layer */
    ds485_client_set_receive_callback(data->client, NULL, NULL);

    pthread_mutex_lock(&data->mutex_callback);
    while (data->callback_list)
    {
        item = data->callback_list;
        HASH_DEL(data->callback_list, item);
        free(item);
    }
    pthread_mutex_unlock(&data->mutex_callback);

    int result = ds485_client_close(data->client);

    data->client = NULL;

    return result;
}

int DsmApiCleanup(DsmApiHandle_t handle)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL || data->client != NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    pthread_mutex_destroy(&data->mutex_callback);
    pthread_mutex_destroy(&data->mutex);
    pthread_cond_destroy(&data->condition);

    free(data);

    return 0;
}

int DsmApiSetCallback(DsmApiHandle_t handle, container_type_t MessageType, uint8_t MessageId, uint8_t ModifierId, DsmApiCallback_t *callback_struct)
{
    struct callback_item *callback;

    uint32_t id = (MessageType << 16) | (MessageId << 8) | (ModifierId);

    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    pthread_mutex_lock(&data->mutex_callback);
    HASH_FIND_INT(data->callback_list, &id, callback);

    if (callback == NULL)
    {
        /* callback not yet in hash */
        if (callback_struct->function != NULL)
        {
            /* create new entry in hash */
            callback = (struct callback_item *)malloc(sizeof(struct callback_item));

            callback->id = id;
            callback->callback = (callback_function_t)callback_struct->function;
            callback->arg = callback_struct->arg;
            HASH_ADD_INT(data->callback_list, id, callback);
            callback_struct->function = NULL;
            callback_struct->arg = NULL;
        }
    }
    else
    {
        /* found callback in hash */
        if (callback_struct->function != NULL)
        {
            DsmApiCallback_t prev_callback;
            prev_callback.function = callback->callback;
            prev_callback.arg = callback->arg;
            /* replace previous entry in hash */
            callback->callback = (callback_function_t)callback_struct->function;
            callback->arg = callback_struct->arg;
            /* no need to add, as it's already part of the hash */
            /* copy previous callback back to user */
            *callback_struct = prev_callback;
        }
        else
        {
            HASH_DEL(data->callback_list, callback);
            /* copy previous callback back to user */
            callback_struct->function = callback->callback;
            callback_struct->arg = callback->arg;
            free(callback);
        }
    }
    pthread_mutex_unlock(&data->mutex_callback);

    return ERROR_OK;
}

int DsmApiSetBusStateCallback(DsmApiHandle_t handle, DsmApiCallback_t *callback_struct)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;
    DsmApiCallback_t prev_cb;

    if (data == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    prev_cb.function = data->bus_state_callback;
    prev_cb.arg = data->bus_state_arg;
    data->bus_state_callback = (BusState_callback_t)callback_struct->function;
    data->bus_state_arg = callback_struct->arg;
    *callback_struct = prev_cb;
    return ERROR_OK;
}

int DsmApiSetBusChangeCallback(DsmApiHandle_t handle, DsmApiCallback_t *callback_struct)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;
    DsmApiCallback_t prev_cb;

    if (data == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    prev_cb.function = data->bus_change_callback;
    prev_cb.arg = data->bus_change_arg;
    data->bus_change_callback = (BusChange_callback_t)callback_struct->function;
    data->bus_change_arg = callback_struct->arg;

    // copy previous struct back to caller
    *callback_struct = prev_cb;
    return ERROR_OK;
}

int DsmApiGetBusMembers(DsmApiHandle_t handle, dsid_t *device_list, int device_list_size)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    return ds485_client_query_devices(data->client, device_list, device_list_size);
}

int DsmApiGetBusState(DsmApiHandle_t handle, bus_state_t *state)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    return ds485_client_query_state(data->client, state);
}

static int dsm_api_ds485_bus_state_callback(void *handle, const bus_state_t state)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return -1;
    }

    if (data->bus_state_callback != NULL)
    {
        data->bus_state_callback(data->bus_state_arg, state);
    }
}

static int dsm_api_ds485_bus_change_callback(void *handle, dsid_t *id, int flags)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return -1;
    }

    if (data->bus_change_callback != NULL)
    {
        data->bus_change_callback(data->bus_change_arg, id, flags);
    }
}

int DsmApiGetOwnDSID(DsmApiHandle_t handle, dsid_t *dsid)
{
    dsm_api_data_t *data = (dsm_api_data_t*)handle;

    if (data == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    return ds485_client_get_dsid(data->client, dsid);
}

int DsmApiExpandDeviceDSID(uint16_t vendorId, uint32_t serialNumber, dsid_t *dsid)
{
    static const dsid_t aizoDevice = {0x35, 0x04, 0x17, 0x5f, 0xe0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
    int retval = 0;

    switch (vendorId)
    {
    case 0x0001:
    {
        *dsid = aizoDevice;
        break;
    }
    default:
    {
        retval = -1;
        memset(dsid->id, 0, sizeof(dsid_t));
        break;
    }
    }

    dsid->id[8] = (serialNumber >> 24) & 0xFF;
    dsid->id[9] = (serialNumber >> 16) & 0xFF;
    dsid->id[10] = (serialNumber >> 8) & 0xFF;
    dsid->id[11] = (serialNumber >> 0) & 0xFF;

    return retval;
}

int DsmApiIsdSM(dsid_t dSMdSID)
{
    if (dSMdSID.id[0] == 0x35 &&
        dSMdSID.id[1] == 0x04 &&
        dSMdSID.id[2] == 0x17 &&
        dSMdSID.id[3] == 0x5f &&
        dSMdSID.id[4] == 0xe0 &&
        dSMdSID.id[5] == 0x00 &&
        dSMdSID.id[6] == 0x00 &&
        dSMdSID.id[7] == 0x10)
    {
        return 1;
    }

    return 0;
}

typedef enum {
    IDLE,
    WAITING,
    DONE,
} CallbackSyncState_t;

typedef struct {
    CallbackSyncState_t state;
    uint16_t device;
    uint8_t offset;
    uint8_t bank;
    uint16_t result;
    uint8_t result2;
} CallbackSync_t;

static void SyncDeviceStatistics_handler(uint8_t error_code, void *arg, dsid_t sourceId, dsid_t destinationId, uint16_t DeviceId, uint8_t Offset, uint16_t Value)
{
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)arg;
    CallbackSync_t *data = (CallbackSync_t*)apiHandle->sync_data;

    if (pthread_mutex_lock(&apiHandle->mutex) == 0)
    {
        if (data != NULL &&
            data->state == WAITING &&
            data->device == DeviceId &&
            data->offset == Offset)
        {
            data->result = Value;
            data->state = DONE;
            pthread_cond_signal(&apiHandle->condition);
        }
        pthread_mutex_unlock(&apiHandle->mutex);
    }
}

int DeviceStatisticsRequest_get_sync(DsmApiHandle_t handle, dsid_t dSMdSID, uint16_t DeviceId, uint8_t ParamOffset, int timeout, uint16_t *Value)
{
    CallbackSync_t *data;
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)handle;
    DsmApiCallback_t callback;
    EventDeviceStatistics_event_callback_t response_callback = &SyncDeviceStatistics_handler;
    void *previousCallback;
    int result;
    struct timespec ts;
    struct timeval now;

    if (handle == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    if (ParamOffset == 0xfe || ParamOffset == 0xff)
    {
        return ERROR_INVALID_PARAMETER;
    }

    data = (CallbackSync_t*)malloc(sizeof(CallbackSync_t));
    apiHandle->sync_data = data;

    pthread_mutex_lock(&apiHandle->mutex);
    data->state = WAITING;
    data->device = DeviceId;
    data->offset = ParamOffset;

    callback.function = (void*)response_callback;
    callback.arg = apiHandle;

    gettimeofday(&now, NULL);
    ts.tv_sec = now.tv_sec + timeout;
    ts.tv_nsec = now.tv_usec * 1000;

    /* replace current callback handler with own */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DEVICE_STATISTICS, 0, &callback);

    /* request statistics value */
    pthread_mutex_unlock(&apiHandle->mutex);
    result = DeviceStatisticsRequest_get(handle, dSMdSID, DeviceId, ParamOffset);
    pthread_mutex_lock(&apiHandle->mutex);
    if (result == 0)
    {
        /* wait for callback */
        if (data->state != DONE) {
            pthread_cond_timedwait(&apiHandle->condition, &apiHandle->mutex, &ts);
        }
        if (data->state == DONE)
        {
            *Value = data->result;
            result = 0;
        } else {
            result = ERROR_RESPONSE_TIMEOUT;
        }
    }
    /* restore callback to previous */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DEVICE_STATISTICS, 0, &callback);
    free(data);
    apiHandle->sync_data = NULL;

    pthread_mutex_unlock(&apiHandle->mutex);

    return result;
}

int DeviceConfig_get_sync_8(DsmApiHandle_t handle, dsid_t dSMdSID, uint16_t DeviceId, uint8_t ConfigClass, uint8_t ConfigIndex, int timeout, uint8_t *Value)
{
    uint16_t value16;
    int result = DeviceConfig_get_sync_16(handle, dSMdSID, DeviceId, ConfigClass, (ConfigIndex & 0xFE), timeout, &value16);
    *Value = ((ConfigIndex & 0x1) == 0) ? (value16 & 0xFF) : (value16 >> 8);
    return result;
}

static void SyncDeviceConfig_16_handler(uint8_t error_code, void *arg, dsid_t sourceId, dsid_t destinationId, uint16_t DeviceId, uint8_t ConfigClass, uint8_t ConfigIndex, uint8_t Value)
{
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)arg;
    CallbackSync_t *data = (CallbackSync_t*)apiHandle->sync_data;
    if (pthread_mutex_lock(&apiHandle->mutex) == 0)
    {
        if (data != NULL &&
            data->state == WAITING &&
            data->device == DeviceId &&
            data->bank == ConfigClass)
        {
            if (data->offset == ConfigIndex)
            {
                data->result += Value;
                data->result2 |= 0x01;
            }
            else if (data->offset + 1 == ConfigIndex)
            {
                data->result += (uint16_t)Value << 8;
                data->result2 |= 0x02;
            }
            if (data->result2 == 0x03)
            {
                data->state = DONE;
                pthread_cond_signal(&apiHandle->condition);
            }
        }
        pthread_mutex_unlock(&apiHandle->mutex);
    }
}

int DeviceConfig_get_sync_16(DsmApiHandle_t handle, dsid_t dSMdSID, uint16_t DeviceId, uint8_t ConfigClass, uint8_t ConfigIndex, int timeout, uint16_t *Value)
{
    CallbackSync_t *data;
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)handle;
    DsmApiCallback_t callback;
    EventDeviceConfig_event_callback_t response_callback = &SyncDeviceConfig_16_handler;
    int result;
    struct timespec ts;
    struct timeval now;

    if (handle == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    if ((ConfigIndex & 0x1) != 0)
    {
        return ERROR_INVALID_PARAMETER;
    }

    data = (CallbackSync_t*)malloc(sizeof(CallbackSync_t));
    apiHandle->sync_data = data;

    pthread_mutex_lock(&apiHandle->mutex);
    data->state = WAITING;
    data->device = DeviceId;
    data->offset = ConfigIndex & 0xFE;
    data->bank = ConfigClass;
    data->result = 0;
    data->result2 = 0;

    callback.function = (void*)response_callback;
    callback.arg = apiHandle;

    gettimeofday(&now, NULL);
    ts.tv_sec = now.tv_sec + timeout;
    ts.tv_nsec = now.tv_usec * 1000;

    /* replace current callback handler with own */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DEVICE_CONFIG, 0, &callback);

    /* request statistics value */
    pthread_mutex_unlock(&apiHandle->mutex);
    result = DeviceConfig_get(handle, dSMdSID, DeviceId, ConfigClass, ConfigIndex);
    pthread_mutex_lock(&apiHandle->mutex);
    if (result == 0)
    {
        /* wait for callback */
        if (data->state != DONE) {
            pthread_cond_timedwait(&apiHandle->condition, &apiHandle->mutex, &ts);
        }
        if (data->state == DONE)
        {
            *Value = data->result;
            result = 0;
        } else {
            result = ERROR_RESPONSE_TIMEOUT;
        }
    }
    /* restore callback to previous */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DEVICE_CONFIG, 0, &callback);
    free(data);
    apiHandle->sync_data = NULL;

    pthread_mutex_unlock(&apiHandle->mutex);

    return result;
}

static void SyncTestTransmissionQuality_handler(uint8_t error_code, void *arg, dsid_t sourceId, dsid_t destinationId, uint16_t DeviceId, uint8_t DownstreamQuality, uint16_t UpstreamQuality)
{
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)arg;
    CallbackSync_t *data = (CallbackSync_t*)apiHandle->sync_data;

    if (pthread_mutex_lock(&apiHandle->mutex) == 0)
    {
        if (data != NULL &&
            data->state == WAITING &&
            data->device == DeviceId)
        {
            data->result = UpstreamQuality;
            data->result2 = DownstreamQuality;
            data->state = DONE;
            pthread_cond_signal(&apiHandle->condition);
        }
        pthread_mutex_unlock(&apiHandle->mutex);
    }
}

int TestTransmissionQuality_get_sync(DsmApiHandle_t handle, dsid_t dSMdSID, uint16_t DeviceId, int timeout, uint8_t *DownstreamQuality, uint16_t *UpstreamQuality)
{
    CallbackSync_t *data;
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)handle;
    DsmApiCallback_t callback;
    EventTestTransmissionQuality_event_callback_t response_callback = &SyncTestTransmissionQuality_handler;
    int result;
    struct timespec ts;
    struct timeval now;

    if (handle == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    data = (CallbackSync_t*)malloc(sizeof(CallbackSync_t));
    apiHandle->sync_data = data;

    pthread_mutex_lock(&apiHandle->mutex);
    data->state = WAITING;
    data->device = DeviceId;

    callback.function = (void*)response_callback;
    callback.arg = apiHandle;

    gettimeofday(&now, NULL);
    ts.tv_sec = now.tv_sec + timeout;
    ts.tv_nsec = now.tv_usec * 1000;

    /* replace current callback handler with own */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_TEST_TRANSMISSION_QUALITY, 0, &callback);

    /* request statistics value */
    pthread_mutex_unlock(&apiHandle->mutex);
    result = TestTransmissionQuality_get(handle, dSMdSID, DeviceId);
    pthread_mutex_lock(&apiHandle->mutex);
    if (result == 0)
    {
        /* wait for callback */
        if (data->state != DONE) {
            pthread_cond_timedwait(&apiHandle->condition, &apiHandle->mutex, &ts);
        }
        if (data->state == DONE)
        {
            *UpstreamQuality = data->result;
            *DownstreamQuality = data->result2;
            result = 0;
        } else {
            result = ERROR_RESPONSE_TIMEOUT;
        }
    }
    /* restore callback to previous */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_TEST_TRANSMISSION_QUALITY, 0, &callback);
    free(data);
    apiHandle->sync_data = NULL;

    pthread_mutex_unlock(&apiHandle->mutex);

    return result;
}

typedef struct {
    CallbackSyncState_t state;
    uint16_t device;
    uint8_t SensorIndex;
    uint16_t SensorValue;
} CallbackSyncSensorValue_t;

static void SyncDeviceSensor_get_handler(uint8_t error_code, void *arg, dsid_t sourceId, dsid_t destinationId, uint16_t DeviceId, uint8_t SensorIndex, uint16_t SensorValue)
{
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)arg;
    CallbackSyncSensorValue_t *data = (CallbackSyncSensorValue_t*)apiHandle->sync_data;

    if (pthread_mutex_lock(&apiHandle->mutex) == 0)
    {
        if (data != NULL &&
            data->state == WAITING &&
            data->device == DeviceId &&
            data->SensorIndex == SensorIndex)
        {
            data->SensorValue = SensorValue;
            data->state = DONE;
            pthread_cond_signal(&apiHandle->condition);
        }
        pthread_mutex_unlock(&apiHandle->mutex);
    }
}

int DeviceSensor_get_value_sync(DsmApiHandle_t handle, dsid_t dSMdSID, uint16_t DeviceId, uint8_t SensorIndex, int timeout, uint16_t *SensorValue)
{
    CallbackSyncSensorValue_t *data;
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)handle;
    DsmApiCallback_t callback;
    EventDeviceSensor_value_event_callback_t response_callback = &SyncDeviceSensor_get_handler;
    int result;
    struct timespec ts;
    struct timeval now;

    if (handle == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    data = (CallbackSyncSensorValue_t*)malloc(sizeof(CallbackSyncSensorValue_t));
    apiHandle->sync_data = data;

    pthread_mutex_lock(&apiHandle->mutex);
    data->state = WAITING;
    data->device = DeviceId;
    data->SensorIndex = SensorIndex;

    callback.function = (void*)response_callback;
    callback.arg = apiHandle;

    gettimeofday(&now, NULL);
    ts.tv_sec = now.tv_sec + timeout;
    ts.tv_nsec = now.tv_usec * 1000;

    /* replace current callback handler with own */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DEVICE_SENSOR, EVENT_DEVICE_SENSOR_VALUE, &callback);

    /* request statistics value */
    pthread_mutex_unlock(&apiHandle->mutex);
    result = DeviceSensor_get_value(handle, dSMdSID, DeviceId, SensorIndex);
    pthread_mutex_lock(&apiHandle->mutex);
    if (result == 0)
    {
        /* wait for callback */
        if (data->state != DONE) {
            pthread_cond_timedwait(&apiHandle->condition, &apiHandle->mutex, &ts);
        }
        if (data->state == DONE)
        {
            *SensorValue = data->SensorValue;
            result = 0;
        } else {
            result = ERROR_RESPONSE_TIMEOUT;
        }
    }
    /* restore callback to previous */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DEVICE_SENSOR, EVENT_DEVICE_SENSOR_VALUE, &callback);
    free(data);
    apiHandle->sync_data = NULL;

    pthread_mutex_unlock(&apiHandle->mutex);

    return result;
}

typedef struct {
    CallbackSyncState_t state;
    uint16_t device;
    uint8_t SensorIndex;
    uint8_t SensorPresent;
    uint8_t SensorType;
    uint8_t LastSensor;
} CallbackSyncSensorType_t;

static void SyncDeviceSensor_type_handler(uint8_t error_code, void *arg, dsid_t sourceId, dsid_t destinationId, uint16_t DeviceId, uint8_t SensorIndex, uint8_t SensorPresent, uint8_t SensorType, uint8_t LastSensor)
{
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)arg;
    CallbackSyncSensorType_t *data = (CallbackSyncSensorType_t*)apiHandle->sync_data;

    if (pthread_mutex_lock(&apiHandle->mutex) == 0)
    {
        if (data != NULL &&
            data->state == WAITING &&
            data->device == DeviceId &&
            data->SensorIndex == SensorIndex)
        {
            data->SensorPresent = SensorPresent;
            data->SensorType = SensorType;
            data->LastSensor = LastSensor;
            data->state = DONE;
            pthread_cond_signal(&apiHandle->condition);
        }
        pthread_mutex_unlock(&apiHandle->mutex);
    }
}

int DeviceSensor_get_type_sync(DsmApiHandle_t handle, dsid_t dSMdSID, uint16_t DeviceId, uint8_t SensorIndex, int timeout, uint8_t *SensorPresent, uint8_t *SensorType, uint8_t *LastSensor)
{
    CallbackSyncSensorType_t *data;
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)handle;
    DsmApiCallback_t callback;
    EventDeviceSensor_type_event_callback_t response_callback = &SyncDeviceSensor_type_handler;
    int result;
    struct timespec ts;
    struct timeval now;

    if (handle == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    data = (CallbackSyncSensorType_t*)malloc(sizeof(CallbackSyncSensorType_t));
    apiHandle->sync_data = data;

    pthread_mutex_lock(&apiHandle->mutex);
    data->state = WAITING;
    data->device = DeviceId;
    data->SensorIndex = SensorIndex;

    callback.function = (void*)response_callback;
    callback.arg = apiHandle;

    gettimeofday(&now, NULL);
    ts.tv_sec = now.tv_sec + timeout;
    ts.tv_nsec = now.tv_usec * 1000;

    /* replace current callback handler with own */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DEVICE_SENSOR, EVENT_DEVICE_SENSOR_TYPE, &callback);

    /* request statistics value */
    pthread_mutex_unlock(&apiHandle->mutex);
    result = DeviceSensor_get_type(handle, dSMdSID, DeviceId, SensorIndex);
    pthread_mutex_lock(&apiHandle->mutex);
    if (result == 0)
    {
        /* wait for callback */
        if (data->state != DONE) {
            pthread_cond_timedwait(&apiHandle->condition, &apiHandle->mutex, &ts);
        }
        if (data->state == DONE)
        {
            *SensorPresent = data->SensorPresent;
            *SensorType = data->SensorType;
            *LastSensor = data->LastSensor;
            result = 0;
        } else {
            result = ERROR_RESPONSE_TIMEOUT;
        }
    }
    /* restore callback to previous */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DEVICE_SENSOR, EVENT_DEVICE_SENSOR_TYPE, &callback);
    free(data);
    apiHandle->sync_data = NULL;

    pthread_mutex_unlock(&apiHandle->mutex);

    return result;
}

static void SyncDSLink_response_handler(uint8_t error_code, void *arg, dsid_t sourceId, dsid_t destinationId, uint16_t DeviceId, uint8_t Value)
{
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)arg;
    CallbackSync_t *data = (CallbackSync_t*)apiHandle->sync_data;

    if (pthread_mutex_lock(&apiHandle->mutex) == 0)
    {
        if (data != NULL &&
            data->state == WAITING &&
            data->device == DeviceId)
        {
            data->result = Value;
            data->state = DONE;
            pthread_cond_signal(&apiHandle->condition);
        }
        pthread_mutex_unlock(&apiHandle->mutex);
    }
}

int DeviceDslinkSend_last_byte_with_answer_sync(DsmApiHandle_t handle, dsid_t dSMdSID, uint16_t DeviceId, uint8_t Value, int timeout, uint8_t *ReturnValue)
{
    CallbackSync_t *data;
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)handle;
    DsmApiCallback_t callback;
    EventDslink_response_event_callback_t response_callback = &SyncDSLink_response_handler;
    int result;
    struct timespec ts;
    struct timeval now;

    if (handle == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    data = (CallbackSync_t*)malloc(sizeof(CallbackSync_t));
    apiHandle->sync_data = data;

    pthread_mutex_lock(&apiHandle->mutex);
    data->state = WAITING;
    data->device = DeviceId;

    callback.function = (void*)response_callback;
    callback.arg = apiHandle;

    gettimeofday(&now, NULL);
    ts.tv_sec = now.tv_sec + timeout;
    ts.tv_nsec = now.tv_usec * 1000;

    /* replace current callback handler with own */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DSLINK, EVENT_DSLINK_RESPONSE, &callback);

    /* request statistics value */
    pthread_mutex_unlock(&apiHandle->mutex);
    result = DeviceDslinkSend_last_byte_with_answer(handle, dSMdSID, DeviceId, Value);
    pthread_mutex_lock(&apiHandle->mutex);
    if (result == 0)
    {
        /* wait for callback */
        if (data->state != DONE) {
            pthread_cond_timedwait(&apiHandle->condition, &apiHandle->mutex, &ts);
        }
        if (data->state == DONE)
        {
            *ReturnValue = (uint8_t)data->result;
            result = 0;
        } else {
            result = ERROR_RESPONSE_TIMEOUT;
        }
    }
    /* restore callback to previous */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DSLINK, EVENT_DSLINK_RESPONSE, &callback);
    free(data);
    apiHandle->sync_data = NULL;

    pthread_mutex_unlock(&apiHandle->mutex);

    return result;
}

int DeviceDslinkSend_not_last_byte_with_answer_sync(DsmApiHandle_t handle, dsid_t dSMdSID, uint16_t DeviceId, uint8_t Value, int timeout, uint8_t *ReturnValue)
{
    CallbackSync_t *data;
    dsm_api_data_t *apiHandle = (dsm_api_data_t*)handle;
    DsmApiCallback_t callback;
    EventDslink_response_event_callback_t response_callback = &SyncDSLink_response_handler;
    int result;
    struct timespec ts;
    struct timeval now;

    if (handle == NULL)
    {
        return ERROR_INVALID_HANDLE;
    }

    data = (CallbackSync_t*)malloc(sizeof(CallbackSync_t));
    apiHandle->sync_data = data;

    pthread_mutex_lock(&apiHandle->mutex);
    data->state = WAITING;
    data->device = DeviceId;

    callback.function = (void*)response_callback;
    callback.arg = data;

    gettimeofday(&now, NULL);
    ts.tv_sec = now.tv_sec + timeout;
    ts.tv_nsec = now.tv_usec * 1000;

    /* replace current callback handler with own */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DSLINK, EVENT_DSLINK_RESPONSE, &callback);

    /* request statistics value */
    pthread_mutex_unlock(&apiHandle->mutex);
    result = DeviceDslinkSend_not_last_byte_with_answer(handle, dSMdSID, DeviceId, Value);
    pthread_mutex_lock(&apiHandle->mutex);
    if (result == 0)
    {
        /* wait for callback */
        if (data->state != DONE) {
            pthread_cond_timedwait(&apiHandle->condition, &apiHandle->mutex, &ts);
        }
        if (data->state == DONE)
        {
            *ReturnValue = (uint8_t)data->result;
            result = 0;
        } else {
            result = ERROR_RESPONSE_TIMEOUT;
        }
    }
    /* restore callback to previous */
    DsmApiSetCallback(handle, DS485_CONTAINER_EVENT, EVENT_DSLINK, EVENT_DSLINK_RESPONSE, &callback);
    free(data);
    apiHandle->sync_data = NULL;

    pthread_mutex_unlock(&apiHandle->mutex);

    return result;
}
